#include <QDate>
#include <QDesktopServices>
#include <QDir>
#include <QDirIterator>
#include <QFileDialog>
#include <QFuture>
#include <QMessageBox>
#include <QProcess>
#include <QScrollBar>
#include <QtConcurrentRun>
#include <QTranslator>

#include "src/qmediainfo.h"

QMediaInfo::QMediaInfo(QWidget *parent) :
  QMainWindow(parent)
{
  loadTranslation();
  setupUi(this);

  if (!MIInstalled()) {
#ifdef Q_OS_WIN
    QMessageBox::warning(this,
                         "Mediainfo",
                         tr("<p>Mediainfo is not installed.</p>"
                         "<p>MediaInfo.exe & MediaInfo.dll must be at the same directory as contain QMediaInfo.exe.</p>"
                         "<p>You can install it from here "
                         "<a href='http://mediainfo.sourceforge.net/'>"
                         "http://mediainfo.sourceforge.net/</a></p>"),
                         QMessageBox::Ok);
#else
    QMessageBox::warning(this,
                         "Mediainfo",
                         tr("<p>Mediainfo is not installed.</p>"
                         "<p>You can install it from here "
                         "<a href='http://mediainfo.sourceforge.net/'>"
                         "http://mediainfo.sourceforge.net/</a></p>"),
                         QMessageBox::Ok);
#endif
    exit(0);
  }

#ifdef Q_OS_WIN
  treeWidget->setFont(QFont("Courier New",10));
  textBrowser->setFont(QFont("Courier New",10));
  setWindowIcon(QIcon("mediainfo.svg"));
#else
  treeWidget->setFont(QFont("Monospace",10));
  textBrowser->setFont(QFont("Monospace",10));
  setWindowIcon(QIcon("/usr/share/icons/hicolor/scalable/mediainfo.svg"));
#endif

  // open/read settings file
  settings = new QSettings(QSettings::NativeFormat, QSettings::UserScope,
                           "qmediainfo", "config");
  loadSettings();

  // all supported containers
  supportedFiles= "*.mkv *.mka *.mks *.ogg *.ogm *.avi *.wav *.mpeg *.mpg *.vob *.mp4 *.rm "
                  "*.mpgv *.mpv *.m1v *.m2v *.mp2 *.mp3 *.asf *.wma *.wmv *.qt *.mov *.aifc "
                  "*.rmvb *.ra *.ifo *.ac3 *.dts *.aac *.ape *.mac *.flac *.dat *.aiff "
                  "*.au *.iff *.paf *.sd2 *.irca *.w64 *.mat *.pvf *.xi *.sds *.avr";

  // load files from cli arguments
  QStringList args = QApplication::arguments();
  args.removeFirst();
  if (!args.isEmpty()) {
    loadFiles(args);
  } else {
    treeWidget->setVisible(false);
    btnAdd->setEnabled(false);
  }

  // needed to use "Remove" in "Hidden items" list
  connect(listHiddenItems,SIGNAL(deletedText(QStringList)),
          this,SLOT(addToComboBox(QStringList)));

  // show info tab on start
  tabWidget->setCurrentIndex(0);
  tabWidget->setTabIcon(0,QIcon(":/mediainfoR.png"));
  tabWidget->setTabIcon(1,QIcon(":/settingsR.png"));

  if (cmbBoxUpdate->currentIndex()!=0)
    checkUpdates();

  // for checkboxes in View menu
  group = new QActionGroup(this);
  group->addAction(actionAsText);
  group->addAction(actionAsTree);
  group->addAction(actionAsPseudoTree);

  connect(actionAsText,      SIGNAL(triggered()),this,SLOT(updateInfo()));
  connect(actionAsTree,      SIGNAL(triggered()),this,SLOT(updateInfo()));
  connect(actionAsPseudoTree,SIGNAL(triggered()),this,SLOT(updateInfo()));

  setAcceptDrops(true);
  resize(800,500);
}

void QMediaInfo::loadTranslation()
{
  if (QLocale::system().name().contains(QRegExp("ru|ua"))) {
    QTranslator *translator = new QTranslator;
    bool loaded = false;
#ifdef Q_OS_WIN
    loaded = translator->load("qmediainfo_ru.qm");
#else
    loaded = translator->load("/usr/share/qmediainfo/qmediainfo_ru");
#endif
    if (loaded)
      qApp->installTranslator(translator);
    else
      QMessageBox::warning(this,
                           "Mediainfo",
                           tr("<p>Can't load translation.</p>"),
                           QMessageBox::Ok);
  }
}

void QMediaInfo::loadSettings()
{
  // settings loading...
  actionAsText->setChecked(settings->value("plainView",true).toBool());
  actionAsTree->setChecked(settings->value("treeView",false).toBool());
  actionAsPseudoTree->setChecked(settings->value("pseudoTree",false).toBool());

  listHiddenItems->addItems(settings->value("hideList").toStringList());

  checkBoxFullInfo->setChecked(settings->value("fullInfo",false).toBool());
  groupBoxHiddenItems->setChecked(settings->value("showHiddenGroupBox",false).toBool());
  chBoxSplitES->setChecked(settings->value("splitES",true).toBool());
  on_chBoxSplitES_stateChanged(chBoxSplitES->checkState());
  cmbBoxUpdate->setCurrentIndex(settings->value("updateTime",2).toInt());
  spinBoxSplit->setValue(settings->value("splitESCount",3).toInt());

  checkBoxWordWrap->setChecked(settings->value("wordWrap",true).toBool());
  if (checkBoxWordWrap->isChecked())
    textBrowser->setWordWrapMode(QTextOption::WordWrap);
  else
    textBrowser->setWordWrapMode(QTextOption::NoWrap);
}

bool QMediaInfo::MIInstalled()
{
#ifdef Q_OS_LINUX
  QProcess availableTest;
  availableTest.start("which", QStringList("mediainfo"));
  availableTest.waitForFinished();
  QString test = availableTest.readAll();
  if (test.contains("which: no mediainfo") || test.isEmpty())
    return false;
  else
    return true;
#endif
#ifdef Q_OS_WIN32
  QFile tempFile("MediaInfo.exe");
  if (tempFile.exists())
    return true;
  else
    return false;
#endif
}

void QMediaInfo::on_openFilesAction_triggered()
{
  QStringList files = QFileDialog::getOpenFileNames(this,
                                         tr("Open File(s)"),
                                         QDir::homePath(),
                                         "Media files ("+supportedFiles+")");
  if (!files.isEmpty())
    loadFiles(files);
}

void QMediaInfo::on_openFolderAction_triggered()
{
  QString selectedDir = QFileDialog::getExistingDirectory(this,
                                                       tr("Select Folder"),
                                                       QDir::homePath(),
                                                       QFileDialog::ShowDirsOnly);
  if (!selectedDir.isNull())
    loadFiles(QStringList(selectedDir));
}

void QMediaInfo::dragEnterEvent(QDragEnterEvent *event)
{
  event->acceptProposedAction();
}

void QMediaInfo::dropEvent(QDropEvent *event)
{
  const QMimeData *mime = event->mimeData();
  if (!mime->hasUrls()) {
    event->ignore();
    return;
  }

  QStringList files;
  for (int i = 0; i<mime->urls().count(); ++i) {
    files.append(mime->urls().at(i).toLocalFile());
  }

  if (!files.isEmpty())
    loadFiles(files);

  event->acceptProposedAction();
}

void QMediaInfo::loadFiles(QStringList files)
{

  QFileInfo tempInfo;
  for (int i = 0; i<files.count(); ++i) {
    tempInfo = QFileInfo(files.at(i));

    // if current path is directory
    if (tempInfo.isDir()) {
      QStringList tempList;
      QDirIterator dirit(tempInfo.absoluteFilePath(),QDirIterator::Subdirectories);
      while (dirit.hasNext()) {
        dirit.next();
        if (QFileInfo(dirit.filePath()).isFile())
          if (!QFileInfo(dirit.filePath()).suffix().isEmpty())
            if (supportedFiles.contains(QFileInfo(dirit.filePath()).suffix()))
              tempList.append(dirit.filePath());
      }

      // sort files from folder
      tempList.sort();
      for (int i = 0; i < tempList.count(); ++i)
        filesList.append(QFileInfo(tempList.at(i)));

    // if current path is file
    } else if (tempInfo.isFile()) {
      if (supportedFiles.contains(tempInfo.suffix()))
        filesList.append(tempInfo);
    }
  }

  // refill combobox
  cmbBoxFiles->clear();
  for (int i = 0; i<filesList.count(); ++i)
    cmbBoxFiles->insertItem(i,filesList.at(i).fileName());
  cmbBoxFiles->setCurrentIndex(filesList.count()-1);
}

QStringList QMediaInfo::launchMI(QString filePath)
{
  //qDebug()<<"MediaInfo launched";

  QString tempStr;
  QStringList arguments;
  QStringList outList;
  QStringList tempList;
  QProcess miProcess;

  if (checkBoxFullInfo->isChecked())
    arguments<<"-f";
  arguments<<filePath;
  miProcess.start("mediainfo", arguments);
  miProcess.waitForFinished();
  tempStr = miProcess.readAllStandardOutput();
#ifdef Q_OS_WIN
  outList = tempStr.split("\r\n");
#else
  outList = tempStr.split("\n");
#endif
  if (groupBoxHiddenItems->isChecked()) {
    for (int i = 0; i<listHiddenItems->count(); ++i) {
      tempList = outList.filter(QRegExp("^"+listHiddenItems->item(i)->text()+" *:"));
      for (int j = 0; j<tempList.count(); ++j)
        outList.removeOne(tempList.at(j));
    }
    fillHiddenItems(outList);
  }

  // remove empty lines at end
  while(outList.last()=="")
    outList.removeLast();

  return outList;
}

void QMediaInfo::fillHiddenItems(const QStringList itemList)
{
  cmbBoxHiddenItems->clear();

  QStringList tempList;
  tempList = settings->value("hideList").toStringList();
  QString pattern;
  for (int i = 0; i<tempList.count(); ++i) {
    pattern.append(tempList.at(i)+"|");
  }

  pattern.remove(QRegExp(".$"));
  QStringList tempList2;
  for (int i = 0; i<itemList.count(); ++i) {
    if (!itemList.at(i).contains(QRegExp("^General|^Video|^Audio|^Text|^Menu"))
        && !itemList.at(i).isEmpty())
      tempList2.append(QString(itemList.at(i)).remove(QRegExp("[ ]*:.*")));
  }

  tempList2.removeDuplicates();
  cmbBoxHiddenItems->addItems(tempList2);
}

void QMediaInfo::on_cmbBoxFiles_currentIndexChanged(int index)
{
  Q_UNUSED(index);

  if (cmbBoxFiles->count()!=0) {
    updateInfo();
    actionCloseAll->setEnabled(true);
    menuView->setEnabled(true);
    menuExport->setEnabled(true);
    actionCloseCurrent->setEnabled(true);
    if (groupBoxHiddenItems->isChecked()) {
      cmbBoxHiddenItems->setEnabled(true);
      btnAdd->setEnabled(true);
    }
  } else {
    actionCloseAll->setEnabled(false);
    menuView->setEnabled(false);
    menuExport->setEnabled(false);
    textBrowser->clear();
    treeWidget->clear();
    cmbBoxHiddenItems->clear();
    actionCloseCurrent->setEnabled(false);
    if (groupBoxHiddenItems->isChecked()) {
      cmbBoxHiddenItems->setEnabled(false);
      btnAdd->setEnabled(false);
    }    
  }
}

void QMediaInfo::updateInfo()
{
  int scrollValue = 0;

  if (!filesList.isEmpty()) {
    setWindowTitle(filesList.at(cmbBoxFiles->currentIndex()).fileName()+" :: QMediaInfo");

    if (actionAsTree->isChecked()) {
      treeWidget->setVisible(true);
      textBrowser->setVisible(false);
      scrollValue = treeWidget->verticalScrollBar()->value();
    } else {
      treeWidget->setVisible(false);
      textBrowser->setVisible(true);
      scrollValue = textBrowser->verticalScrollBar()->value();
    }

    // create additional thread to read mediainfo output
    QFuture<QStringList> future = QtConcurrent::run(this, &QMediaInfo::launchMI,
                                  filesList.at(cmbBoxFiles->currentIndex()).filePath());

    if (actionAsText->isChecked()) {
      fillTextBrowser(future.result());
      textBrowser->verticalScrollBar()->setValue(scrollValue);
    }

    if (actionAsTree->isChecked()) {
      fillTree(future.result());
      treeWidget->verticalScrollBar()->setValue(scrollValue);
    }

    if (actionAsPseudoTree->isChecked()) {
      fillPseudoTree(future.result());
      textBrowser->verticalScrollBar()->setValue(scrollValue);
    }
  }
}

void QMediaInfo::fillTextBrowser(const QStringList infoList)
{
  textBrowser->clear();

  QString tmpStr;
  if (checkBoxWordWrap->isChecked())
    textBrowser->setWordWrapMode(QTextOption::WordWrap);
  for (int i = 0; i<infoList.count(); ++i)
    tmpStr.append(infoList.at(i)+"\n");

  textBrowser->setText(tmpStr);
}

void QMediaInfo::fillTree(const QStringList infoList)
{
  treeWidget->clear();

  bool prevEmpty = true;
  QStringList tempList = infoList;
  QTreeWidgetItem *main;
  QTreeWidgetItem *child;

  for (int i = 0; i < tempList.count(); ++i) {
    if (prevEmpty) {
      main = new QTreeWidgetItem(treeWidget);
      main->setText(0,tempList.at(i));
      prevEmpty = false;
    } else if (tempList.at(i)!="") {
      child = new QTreeWidgetItem(main);
      //child->setFlags(Qt::ItemIsEditable | Qt::ItemIsEnabled);
      if (tempList.at(i).contains("Encoding settings")) {
        QLabel *tempLabel = new QLabel(QString(tempList.at(i)));
        tempLabel->setWordWrap(true);
        tempLabel->setContentsMargins(3,0,0,0);
        treeWidget->setItemWidget(child,0,tempLabel);
      } else
        child->setText(0,tempList.at(i));
    } else {
      prevEmpty = true;
    }
  }

  treeWidget->expandAll();
}

void QMediaInfo::fillPseudoTree(const QStringList infoList)
{
  textBrowser->clear();

  bool prevEmpty = true;
  QString tmpStr;
  QStringList tempList = infoList;

  textBrowser->setWordWrapMode(QTextOption::NoWrap);

  for (int i = 0; i < tempList.count(); ++i) {
    if (prevEmpty) {
      tmpStr.append("<span style=\" font-size:10pt; font-weight:600;\">"
                    +tempList.at(i)+"</span><br />");
      prevEmpty = false;
    } else if (tempList.at(i)!="") {
      if (chBoxSplitES->isChecked() && tempList.at(i).contains("Encoding settings")) {
        tmpStr.append("-- Encoding settings");
        QStringList tempList2;
        tempList2 = QString(QString(tempList.at(i)).remove(QRegExp("^.*: "))).split("/");
        tempList2.prepend("");
        for (int i = 1; i<tempList2.count(); ++i) {
          if (i==1) {
            for (int i = 0; i<15; ++i)
              tmpStr.append("&nbsp;");
            tmpStr.append(" : ");
          }
          tmpStr.append(tempList2.at(i));

          if (i!=tempList2.count()-1)
            tmpStr.append(" / ");

          if (i%spinBoxSplit->value()==0 && i!=0 && i!=tempList2.count()-1) {
            tmpStr.append("<br />");
            for (int i = 0; i<37; ++i)
              tmpStr.append("&nbsp;");
          }
        }
        tmpStr.append("<br />");
      } else {
        tmpStr.append("-- "+QString(tempList.at(i)).replace(" ","&nbsp;")+"<br />");
      }
    } else {
      prevEmpty = true;
    }
  }

  textBrowser->setText(tmpStr);
}

void QMediaInfo::closeEvent(QCloseEvent *)
{
  // saving settings
  QStringList tempList;
  for (int i = 0; i<listHiddenItems->count(); ++i)
    tempList.append(listHiddenItems->item(i)->text());
  settings->setValue("hideList",tempList);

  tempList.clear();
  for (int i = 0; i<cmbBoxHiddenItems->count(); ++i)
    tempList.append(cmbBoxHiddenItems->itemText(i));
  settings->setValue("defaultHiddenItemsList",tempList);

  settings->setValue("fullInfo",checkBoxFullInfo->isChecked());
  settings->setValue("showHiddenGroupBox",groupBoxHiddenItems->isChecked());
  settings->setValue("plainView",actionAsText->isChecked());
  settings->setValue("treeView",actionAsTree->isChecked());
  settings->setValue("pseudoTree",actionAsPseudoTree->isChecked());
  settings->setValue("wordWrap",checkBoxWordWrap->isChecked());
  settings->setValue("updateTime",cmbBoxUpdate->currentIndex());
  settings->setValue("splitES",chBoxSplitES->isChecked());
  settings->setValue("splitESCount",spinBoxSplit->value());
}

void QMediaInfo::on_actionToPlain_triggered()
{
  QString fileName =
      QFileDialog::getSaveFileName(this, tr("Save File"),
                                   QDir::homePath()+QDir::separator()+cmbBoxFiles->currentText(),
                                  "Text files (*.txt)");
  exportInfo(fileName,"plain");
}

void QMediaInfo::on_actionToHTML_triggered()
{
  QString fileName =
      QFileDialog::getSaveFileName(this, tr("Save File"),
                                   QDir::homePath()+QDir::separator()+cmbBoxFiles->currentText(),
                                  "HTML files (*.html)");
  exportInfo(fileName,"HTML");
}

void QMediaInfo::on_actionToXML_triggered()
{
  QString fileName =
      QFileDialog::getSaveFileName(this, tr("Save File"),
                                   QDir::homePath()+QDir::separator()+cmbBoxFiles->currentText(),
                                  "XML files (*.xml)");
  exportInfo(fileName,"XML");
}

// export info to a file
void QMediaInfo::exportInfo(QString file, QString type)
{
  QProcess exp;
  QStringList arguments;
  arguments.append(filesList.at(cmbBoxFiles->currentIndex()).absoluteFilePath());
  if (type=="XML")
    arguments.append("--Output=XML");
  if (type=="HTML")
    arguments.append("--Output=HTML");
  exp.start("mediainfo",arguments);
  exp.waitForFinished();

  QFile outFile(file);
  if (!outFile.open(QIODevice::WriteOnly | QIODevice::Text))
    return;
  QTextStream out(&outFile);
#ifdef Q_OS_WIN
  out.setCodec("Windows-1251");
#else
  out.setCodec("UTF-8");
#endif
  out<<exp.readAllStandardOutput();
  outFile.close();
}

void QMediaInfo::on_actionCloseAll_triggered()
{
  cmbBoxFiles->clear();
  filesList.clear();
  setWindowTitle("QMediaInfo");
}

void QMediaInfo::on_actionCloseCurrent_triggered()
{
  cmbBoxFiles->removeItem(cmbBoxFiles->currentIndex());
  filesList.removeAt(cmbBoxFiles->currentIndex()+1);
}

void QMediaInfo::on_checkBoxWordWrap_stateChanged(int value)
{
  if (value==Qt::Checked)
    textBrowser->setWordWrapMode(QTextOption::WordWrap);
  else
    textBrowser->setWordWrapMode(QTextOption::NoWrap);
}

void QMediaInfo::on_btnAdd_clicked()
{
  if (!cmbBoxHiddenItems->currentText().isEmpty()) {
    listHiddenItems->addItem(cmbBoxHiddenItems->currentText());
    cmbBoxHiddenItems->removeItem(cmbBoxHiddenItems->currentIndex());
    updateInfo();
  }

  if (cmbBoxHiddenItems->count() > 0)
    btnAdd->setEnabled(true);
  else
    btnAdd->setEnabled(false);
}

void QMediaInfo::on_checkBoxFullInfo_stateChanged(int value)
{
  Q_UNUSED(value);
  if (!filesList.isEmpty())
    updateInfo();
}

void QMediaInfo::on_groupBoxHiddenItems_clicked(bool checked)
{
  Q_UNUSED(checked);
  if (!filesList.isEmpty())
    updateInfo();
}

void QMediaInfo::addToComboBox(const QStringList itemList)
{
  for (int i = 0; i<itemList.count(); ++i) {
    cmbBoxHiddenItems->addItem(itemList.at(i));
  }

  updateInfo();
}

void QMediaInfo::on_actionAbout_triggered()
{
  QMessageBox *about = new QMessageBox();
  about->about(this,tr("About"),tr("Simple MediaInfo parcer based on Qt library.\nVersion 0.3.1"));
}

void QMediaInfo::on_actionAboutQt_triggered()
{
  QMessageBox *aboutQt = new QMessageBox();
  aboutQt->aboutQt(this,tr("About Qt"));
}

void QMediaInfo::on_actionAboutMediaInfo_triggered()
{
  QMessageBox *aboutMediaInfo = new QMessageBox();
  QProcess verProcess;
  verProcess.start("mediainfo", QStringList("--Version"));
  verProcess.waitForFinished();
  aboutMediaInfo->about(this,tr("About MediaInfo"),
                        "Copyright (C) 2002-2011 MediaArea.net SARL\n"+verProcess.readAll());
}

void QMediaInfo::on_actionSupportedFiles_triggered()
{
  QMessageBox *aboutSupportedFiles = new QMessageBox();
  aboutSupportedFiles->about(this,tr("Supported Files"),supportedFiles);
}

void QMediaInfo::checkUpdates()
{
  int dayOfYear = QDate::currentDate().dayOfYear();
  int checkDays=0;
  switch (settings->value("updateTime",2).toInt()) {
    case 1 : checkDays = 1;  break;
    case 2 : checkDays = 7;  break;
    case 3 : checkDays = 30; break;
  }

  if (!settings->value("lastUpdCheck").isNull()) {
    if (settings->value("lastUpdCheck",0).toInt()+checkDays <= dayOfYear) {
      QNetworkAccessManager *site = new QNetworkAccessManager(this);
      connect(site, SIGNAL(finished(QNetworkReply*)),
              this, SLOT(replyFinished(QNetworkReply*)));
      site->get(QNetworkRequest(QUrl("http://mediainfo.sourceforge.net/en/Download/Windows")));
      settings->setValue("lastUpdCheck",dayOfYear);
    }
  } else {
    settings->setValue("lastUpdCheck",dayOfYear);
  }
}

void QMediaInfo::replyFinished(QNetworkReply* reply)
{
  QString source = reply->readAll();
  QStringList sourceList = source.split("\n");
  QString verFromNet = sourceList.filter(
                    "http://sourceforge.net/projects/mediainfo/files/binary/mediainfo").at(0);
  verFromNet.replace(QRegExp(".*gui/|/MediaInfo_GUI.*"),"");
  QStringList getVerNet = verFromNet.split(".");

  QProcess verFromCli;
  verFromCli.start("mediainfo", QStringList("--Version"));
  verFromCli.waitForFinished();
  QStringList getVerCli = QString(verFromCli.readAll()).replace(QRegExp(".*v|\n"),"").split(".");

  int newVer=0;
  int nowVer=0;
  for (int i = 0; i<getVerNet.count(); ++i) {
    if (getVerNet.at(i).toInt()!=0)
      newVer+=getVerNet.at(i).toInt();
    if (getVerCli.at(i).toInt()!=0)
      nowVer+=getVerCli.at(i).toInt();
  }

  if (newVer>nowVer) {
    int ans = QMessageBox::information(this,tr("QMediaInfo Update"),
                                       tr("New version of MediaInfo available.\n"
                                       "Download it now?"),
                                       QMessageBox::Yes,QMessageBox::No);
    if (ans==QMessageBox::Yes)
      QDesktopServices::openUrl(QUrl("http://mediainfo.sourceforge.net/"));
  }
}

void QMediaInfo::on_actionExit_triggered()
{
  close();
}

void QMediaInfo::on_chBoxSplitES_stateChanged(int value)
{
  if (value==Qt::Unchecked) {
    lblSplitOpt->setEnabled(false);
    spinBoxSplit->setEnabled(false);
  } else {
    lblSplitOpt->setEnabled(true);
    spinBoxSplit->setEnabled(true);
  }

  updateInfo();
}

void QMediaInfo::on_spinBoxSplit_valueChanged(int value)
{
  Q_UNUSED(value)
  updateInfo();
}

void QMediaInfo::on_groupBoxHiddenItems_toggled(bool value)
{
  Q_UNUSED (value)
  updateInfo();
}
